# --------------------------------------------------------------------------------------
# Copyright (c) 2014-2021, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# --------------------------------------------------------------------------------------
import pytest

from kiwisolver import strength


def test_accessing_predefined_strength() -> None:
    """Test getting the default values for the strength."""
    assert strength.weak < strength.medium
    assert strength.medium < strength.strong
    assert strength.strong < strength.required


def test_creating_strength() -> None:
    """Test creating strength from constituent values."""
    assert strength.create(0, 0, 1) < strength.create(0, 1, 0)
    assert strength.create(0, 1, 0) < strength.create(1, 0, 0)
    assert strength.create(1, 0, 0, 1) < strength.create(1, 0, 0, 4)

    with pytest.raises(TypeError):
        strength.create("", "", "")  # type: ignore



# --------------------------------------------------------------------------------------
# Copyright (c) 2019, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# --------------------------------------------------------------------------------------
"""Time updating an EditVariable in a set of constraints typical of enaml use.

"""
import perf

from kiwisolver import Solver, Variable, strength

solver = Solver()

# Create custom strength
mmedium = strength.create(0, 1, 0, 1.25)
smedium = strength.create(0, 100, 0)

# Create the variable
left = Variable("left")
height = Variable("height")
top = Variable("top")
width = Variable("width")
contents_top = Variable("contents_top")
contents_bottom = Variable("contents_bottom")
contents_left = Variable("contents_left")
contents_right = Variable("contents_right")
midline = Variable("midline")
ctleft = Variable("ctleft")
ctheight = Variable("ctheight")
cttop = Variable("cttop")
ctwidth = Variable("ctwidth")
lb1left = Variable("lb1left")
lb1height = Variable("lb1height")
lb1top = Variable("lb1top")
lb1width = Variable("lb1width")
lb2left = Variable("lb2left")
lb2height = Variable("lb2height")
lb2top = Variable("lb2top")
lb2width = Variable("lb2width")
lb3left = Variable("lb3left")
lb3height = Variable("lb3height")
lb3top = Variable("lb3top")
lb3width = Variable("lb3width")
fl1left = Variable("fl1left")
fl1height = Variable("fl1height")
fl1top = Variable("fl1top")
fl1width = Variable("fl1width")
fl2left = Variable("fl2left")
fl2height = Variable("fl2height")
fl2top = Variable("fl2top")
fl2width = Variable("fl2width")
fl3left = Variable("fl3left")
fl3height = Variable("fl3height")
fl3top = Variable("fl3top")
fl3width = Variable("fl3width")

# Add the edit variables
solver.addEditVariable(width, "strong")
solver.addEditVariable(height, "strong")

# Add the constraints
for c in [
    (left + -0 >= 0) | "required",
    (height + 0 == 0) | "medium",
    (top + -0 >= 0) | "required",
    (width + -0 >= 0) | "required",
    (height + -0 >= 0) | "required",
    (-top + contents_top + -10 == 0) | "required",
    (lb3height + -16 == 0) | "strong",
    (lb3height + -16 >= 0) | "strong",
    (ctleft + -0 >= 0) | "required",
    (cttop + -0 >= 0) | "required",
    (ctwidth + -0 >= 0) | "required",
    (ctheight + -0 >= 0) | "required",
    (fl3left + -0 >= 0) | "required",
    (ctheight + -24 >= 0) | smedium,
    (ctwidth + -1.67772e07 <= 0) | smedium,
    (ctheight + -24 <= 0) | smedium,
    (fl3top + -0 >= 0) | "required",
    (fl3width + -0 >= 0) | "required",
    (fl3height + -0 >= 0) | "required",
    (lb1width + -67 == 0) | "weak",
    (lb2width + -0 >= 0) | "required",
    (lb2height + -0 >= 0) | "required",
    (fl2height + -0 >= 0) | "required",
    (lb3left + -0 >= 0) | "required",
    (fl2width + -125 >= 0) | "strong",
    (fl2height + -21 == 0) | "strong",
    (fl2height + -21 >= 0) | "strong",
    (lb3top + -0 >= 0) | "required",
    (lb3width + -0 >= 0) | "required",
    (fl1left + -0 >= 0) | "required",
    (fl1width + -0 >= 0) | "required",
    (lb1width + -67 >= 0) | "strong",
    (fl2left + -0 >= 0) | "required",
    (lb2width + -66 == 0) | "weak",
    (lb2width + -66 >= 0) | "strong",
    (lb2height + -16 == 0) | "strong",
    (fl1height + -0 >= 0) | "required",
    (fl1top + -0 >= 0) | "required",
    (lb2top + -0 >= 0) | "required",
    (-lb2top + lb3top + -lb2height + -10 == 0) | mmedium,
    (-lb3top + -lb3height + fl3top + -10 >= 0) | "required",
    (-lb3top + -lb3height + fl3top + -10 == 0) | mmedium,
    (contents_bottom + -fl3height + -fl3top + -0 == 0) | mmedium,
    (fl1top + -contents_top + 0 >= 0) | "required",
    (fl1top + -contents_top + 0 == 0) | mmedium,
    (contents_bottom + -fl3height + -fl3top + -0 >= 0) | "required",
    (-left + -width + contents_right + 10 == 0) | "required",
    (-top + -height + contents_bottom + 10 == 0) | "required",
    (-left + contents_left + -10 == 0) | "required",
    (lb3left + -contents_left + 0 == 0) | mmedium,
    (fl1left + -midline + 0 == 0) | "strong",
    (fl2left + -midline + 0 == 0) | "strong",
    (ctleft + -midline + 0 == 0) | "strong",
    (fl1top + 0.5 * fl1height + -lb1top + -0.5 * lb1height + 0 == 0) | "strong",
    (lb1left + -contents_left + 0 >= 0) | "required",
    (lb1left + -contents_left + 0 == 0) | mmedium,
    (-lb1left + fl1left + -lb1width + -10 >= 0) | "required",
    (-lb1left + fl1left + -lb1width + -10 == 0) | mmedium,
    (-fl1left + contents_right + -fl1width + -0 >= 0) | "required",
    (width + 0 == 0) | "medium",
    (-fl1top + fl2top + -fl1height + -10 >= 0) | "required",
    (-fl1top + fl2top + -fl1height + -10 == 0) | mmedium,
    (cttop + -fl2top + -fl2height + -10 >= 0) | "required",
    (-ctheight + -cttop + fl3top + -10 >= 0) | "required",
    (contents_bottom + -fl3height + -fl3top + -0 >= 0) | "required",
    (cttop + -fl2top + -fl2height + -10 == 0) | mmedium,
    (-fl1left + contents_right + -fl1width + -0 == 0) | mmedium,
    (-lb2top + -0.5 * lb2height + fl2top + 0.5 * fl2height + 0 == 0) | "strong",
    (-contents_left + lb2left + 0 >= 0) | "required",
    (-contents_left + lb2left + 0 == 0) | mmedium,
    (fl2left + -lb2width + -lb2left + -10 >= 0) | "required",
    (-ctheight + -cttop + fl3top + -10 == 0) | mmedium,
    (contents_bottom + -fl3height + -fl3top + -0 == 0) | mmedium,
    (lb1top + -0 >= 0) | "required",
    (lb1width + -0 >= 0) | "required",
    (lb1height + -0 >= 0) | "required",
    (fl2left + -lb2width + -lb2left + -10 == 0) | mmedium,
    (-fl2left + -fl2width + contents_right + -0 == 0) | mmedium,
    (-fl2left + -fl2width + contents_right + -0 >= 0) | "required",
    (lb3left + -contents_left + 0 >= 0) | "required",
    (lb1left + -0 >= 0) | "required",
    (0.5 * ctheight + cttop + -lb3top + -0.5 * lb3height + 0 == 0) | "strong",
    (ctleft + -lb3left + -lb3width + -10 >= 0) | "required",
    (-ctwidth + -ctleft + contents_right + -0 >= 0) | "required",
    (ctleft + -lb3left + -lb3width + -10 == 0) | mmedium,
    (fl3left + -contents_left + 0 >= 0) | "required",
    (fl3left + -contents_left + 0 == 0) | mmedium,
    (-ctwidth + -ctleft + contents_right + -0 == 0) | mmedium,
    (-fl3left + contents_right + -fl3width + -0 == 0) | mmedium,
    (-contents_top + lb1top + 0 >= 0) | "required",
    (-contents_top + lb1top + 0 == 0) | mmedium,
    (-fl3left + contents_right + -fl3width + -0 >= 0) | "required",
    (lb2top + -lb1top + -lb1height + -10 >= 0) | "required",
    (-lb2top + lb3top + -lb2height + -10 >= 0) | "required",
    (lb2top + -lb1top + -lb1height + -10 == 0) | mmedium,
    (fl1height + -21 == 0) | "strong",
    (fl1height + -21 >= 0) | "strong",
    (lb2left + -0 >= 0) | "required",
    (lb2height + -16 >= 0) | "strong",
    (fl2top + -0 >= 0) | "required",
    (fl2width + -0 >= 0) | "required",
    (lb1height + -16 >= 0) | "strong",
    (lb1height + -16 == 0) | "strong",
    (fl3width + -125 >= 0) | "strong",
    (fl3height + -21 == 0) | "strong",
    (fl3height + -21 >= 0) | "strong",
    (lb3height + -0 >= 0) | "required",
    (ctwidth + -119 >= 0) | smedium,
    (lb3width + -24 == 0) | "weak",
    (lb3width + -24 >= 0) | "strong",
    (fl1width + -125 >= 0) | "strong",
]:
    solver.addConstraint(c)


def bench_update_variables(loops, solver):
    """Suggest new values and update variables.

    This mimic the use of kiwi in enaml in the case of a resizing.

    """
    t0 = perf.perf_counter()
    for w, h in [
        (400, 600),
        (600, 400),
        (800, 1200),
        (1200, 800),
        (400, 800),
        (800, 400),
    ] * loops:
        solver.suggestValue(width, w)
        solver.suggestValue(height, h)
        solver.updateVariables()

    return perf.perf_counter() - t0


runner = perf.Runner()
runner.bench_time_func(
    "kiwi.suggestValue", bench_update_variables, solver, inner_loops=1
)



# --------------------------------------------------------------------------------------
# Copyright (c) 2013-2022, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# --------------------------------------------------------------------------------------
from ._cext import (
    BadRequiredStrength,
    Constraint,
    DuplicateConstraint,
    DuplicateEditVariable,
    Expression,
    Solver,
    Term,
    UnknownConstraint,
    UnknownEditVariable,
    UnsatisfiableConstraint,
    Variable,
    __kiwi_version__,
    __version__,
    strength,
)

__all__ = [
    "BadRequiredStrength",
    "DuplicateConstraint",
    "DuplicateEditVariable",
    "UnknownConstraint",
    "UnknownEditVariable",
    "UnsatisfiableConstraint",
    "strength",
    "Variable",
    "Term",
    "Expression",
    "Constraint",
    "Solver",
    "__version__",
    "__kiwi_version__",
]



# --------------------------------------------------------------------------------------
# Copyright (c) 2014-2022, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# --------------------------------------------------------------------------------------
import pytest

from kiwisolver import (
    BadRequiredStrength,
    DuplicateConstraint,
    DuplicateEditVariable,
    Solver,
    UnknownConstraint,
    UnknownEditVariable,
    UnsatisfiableConstraint,
    Variable,
)


def test_solver_creation() -> None:
    """Test initializing a solver."""
    s = Solver()
    assert isinstance(s, Solver)

    with pytest.raises(TypeError):
        Solver(Variable())  # type: ignore


def test_managing_edit_variable() -> None:
    """Test adding/removing edit variables."""
    s = Solver()
    v1 = Variable("foo")
    v2 = Variable("bar")

    with pytest.raises(TypeError):
        s.hasEditVariable(object())  # type: ignore
    with pytest.raises(TypeError):
        s.addEditVariable(object(), "weak")  # type: ignore
    with pytest.raises(TypeError):
        s.removeEditVariable(object())  # type: ignore
    with pytest.raises(TypeError):
        s.suggestValue(object(), 10)  # type: ignore

    assert not s.hasEditVariable(v1)
    s.addEditVariable(v1, "weak")
    assert s.hasEditVariable(v1)
    with pytest.raises(DuplicateEditVariable):
        s.addEditVariable(v1, "medium")
    with pytest.raises(UnknownEditVariable):
        s.removeEditVariable(v2)
    s.removeEditVariable(v1)
    assert not s.hasEditVariable(v1)

    with pytest.raises(BadRequiredStrength):
        s.addEditVariable(v1, "required")

    s.addEditVariable(v2, "strong")
    assert s.hasEditVariable(v2)
    with pytest.raises(UnknownEditVariable):
        s.suggestValue(v1, 10)

    s.reset()
    assert not s.hasEditVariable(v2)


def test_suggesting_values_for_edit_variables() -> None:
    """Test suggesting values in different situations."""
    # Suggest value for an edit variable entering a weak equality
    s = Solver()
    v1 = Variable("foo")

    s.addEditVariable(v1, "medium")
    s.addConstraint((v1 == 1) | "weak")
    s.suggestValue(v1, 2)
    s.updateVariables()
    assert v1.value() == 2

    # Suggest a value for an edit variable entering multiple solver rows
    s.reset()
    v1 = Variable("foo")
    v2 = Variable("bar")
    s = Solver()

    s.addEditVariable(v2, "weak")
    s.addConstraint(v1 + v2 == 0)
    s.addConstraint((v2 <= -1))
    s.addConstraint((v2 >= 0) | "weak")
    s.suggestValue(v2, 0)
    s.updateVariables()
    assert v2.value() <= -1


def test_managing_constraints() -> None:
    """Test adding/removing constraints."""
    s = Solver()
    v = Variable("foo")
    c1 = v >= 1
    c2 = v <= 0

    with pytest.raises(TypeError):
        s.hasConstraint(object())  # type: ignore
    with pytest.raises(TypeError):
        s.addConstraint(object())  # type: ignore
    with pytest.raises(TypeError):
        s.removeConstraint(object())  # type: ignore

    assert not s.hasConstraint(c1)
    s.addConstraint(c1)
    assert s.hasConstraint(c1)
    with pytest.raises(DuplicateConstraint):
        s.addConstraint(c1)
    with pytest.raises(UnknownConstraint):
        s.removeConstraint(c2)
    with pytest.raises(UnsatisfiableConstraint):
        s.addConstraint(c2)
    # XXX need to find how to get an invalid symbol from choose subject
    # with pytest.raises(UnsatisfiableConstraint):
    #     s.addConstraint(c3)
    s.removeConstraint(c1)
    assert not s.hasConstraint(c1)

    s.addConstraint(c2)
    assert s.hasConstraint(c2)
    s.reset()
    assert not s.hasConstraint(c2)


def test_solving_under_constrained_system() -> None:
    """Test solving an under constrained system."""
    s = Solver()
    v = Variable("foo")
    c = 2 * v + 1 >= 0
    s.addEditVariable(v, "weak")
    s.addConstraint(c)
    s.suggestValue(v, 10)
    s.updateVariables()

    assert c.expression().value() == 21
    assert c.expression().terms()[0].value() == 20
    assert c.expression().terms()[0].variable().value() == 10


def test_solving_with_strength() -> None:
    """Test solving a system with unsatisfiable non-required constraint."""
    v1 = Variable("foo")
    v2 = Variable("bar")
    s = Solver()

    s.addConstraint(v1 + v2 == 0)
    s.addConstraint(v1 == 10)
    s.addConstraint((v2 >= 0) | "weak")
    s.updateVariables()
    assert v1.value() == 10 and v2.value() == -10

    s.reset()

    s.addConstraint(v1 + v2 == 0)
    s.addConstraint((v1 >= 10) | "medium")
    s.addConstraint((v2 == 2) | "strong")
    s.updateVariables()
    assert v1.value() == -2 and v2.value() == 2


# Typical output solver.dump in the following function.
# the order is not stable.
# """Objective
# ---------
# -2 + 2 * e2 + 1 * s8 + -2 * s10

# Tableau
# -------
# v1 | 1 + 1 * s10
# e3 | -1 + 1 * e2 + -1 * s10
# v4 | -1 + -1 * d5 + -1 * s10
# s6 | -2 + -1 * s10
# e9 | -1 + 1 * s8 + -1 * s10

# Infeasible
# ----------
# e3
# e9

# Variables
# ---------
# bar = v1
# foo = v4

# Edit Variables
# --------------
# bar

# Constraints
# -----------
# 1 * bar + -0 >= 0  | strength = 1
# 1 * bar + 1 <= 0  | strength = 1.001e+09
# 1 * foo + 1 * bar + 0 == 0  | strength = 1.001e+09
# 1 * bar + 0 == 0  | strength = 1

# """


def test_dumping_solver(capsys) -> None:
    """Test dumping the solver internal to stdout."""
    v1 = Variable("foo")
    v2 = Variable("bar")
    s = Solver()

    s.addEditVariable(v2, "weak")
    s.addConstraint(v1 + v2 == 0)
    s.addConstraint((v2 <= -1))
    s.addConstraint((v2 >= 0) | "weak")
    s.updateVariables()
    try:
        s.addConstraint((v2 >= 1))
    except Exception:
        pass

    # Print the solver state
    s.dump()

    state = s.dumps()
    for header in (
        "Objective",
        "Tableau",
        "Infeasible",
        "Variables",
        "Edit Variables",
        "Constraints",
    ):
        assert header in state


def test_handling_infeasible_constraints() -> None:
    """Test that we properly handle infeasible constraints.

    We use the example of the cassowary paper to generate an infeasible
    situation after updating an edit variable which causes the solver to use
    the dual optimization.

    """
    xm = Variable("xm")
    xl = Variable("xl")
    xr = Variable("xr")
    s = Solver()

    s.addEditVariable(xm, "strong")
    s.addEditVariable(xl, "weak")
    s.addEditVariable(xr, "weak")
    s.addConstraint(2 * xm == xl + xr)
    s.addConstraint(xl + 20 <= xr)
    s.addConstraint(xl >= -10)
    s.addConstraint(xr <= 100)

    s.suggestValue(xm, 40)
    s.suggestValue(xr, 50)
    s.suggestValue(xl, 30)

    # First update causing a normal update.
    s.suggestValue(xm, 60)

    # Create an infeasible condition triggering a dual optimization
    s.suggestValue(xm, 90)
    s.updateVariables()
    assert xl.value() + xr.value() == 2 * xm.value()
    assert xl.value() == 80
    assert xr.value() == 100


def test_constraint_violated():
    """Test running a solver and check that constraints
    report they've been violated

    """
    s = Solver()
    v = Variable("foo")

    c1 = (v >= 10) | "required"
    c2 = (v <= -5) | "weak"

    s.addConstraint(c1)
    s.addConstraint(c2)

    s.updateVariables()

    assert v.value() >= 10
    assert c1.violated() is False
    assert c2.violated() is True



# --------------------------------------------------------------------------------------
# Copyright (c) 2014-2021, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# --------------------------------------------------------------------------------------
import gc
import math
import operator
from typing import Tuple

import pytest

from kiwisolver import Constraint, Expression, Term, Variable, strength


def test_expression_creation() -> None:
    """Test the Term constructor."""
    v = Variable("foo")
    v2 = Variable("bar")
    v3 = Variable("aux")
    e1 = Expression((v * 1, v2 * 2, v3 * 3))
    e2 = Expression((v * 1, v2 * 2, v3 * 3), 10)

    for e, val in ((e1, 0), (e2, 10)):
        t = e.terms()
        assert (
            len(t) == 3
            and t[0].variable() is v
            and t[0].coefficient() == 1
            and t[1].variable() is v2
            and t[1].coefficient() == 2
            and t[2].variable() is v3
            and t[2].coefficient() == 3
        )
        assert e.constant() == val

    assert str(e2) == "1 * foo + 2 * bar + 3 * aux + 10"

    with pytest.raises(TypeError) as excinfo:
        Expression((1, v2 * 2, v3 * 3))  # type: ignore
    assert "Term" in excinfo.exconly()

    # ensure we test garbage collection.
    del e2
    gc.collect()


@pytest.fixture()
def expressions():
    """Build expressions, terms and variables to test operations."""
    v = Variable("foo")
    v2 = Variable("bar")
    t = Term(v, 10)
    t2 = Term(v2)
    e = t + 5
    e2 = v2 - 10
    return e, e2, t, t2, v, v2


def test_expression_neg(
    expressions: Tuple[Expression, Expression, Term, Term, Variable, Variable]
):
    """Test neg on an expression."""
    e, _, _, _, v, _ = expressions

    neg = -e
    assert isinstance(neg, Expression)
    neg_t = neg.terms()
    assert (
        len(neg_t) == 1
        and neg_t[0].variable() is v
        and neg_t[0].coefficient() == -10
        and neg.constant() == -5
    )


def test_expression_mul(
    expressions: Tuple[Expression, Expression, Term, Term, Variable, Variable]
):
    """Test expresion multiplication."""
    e, _, _, _, v, _ = expressions

    for mul in (e * 2.0, 2.0 * e):
        assert isinstance(mul, Expression)
        mul_t = mul.terms()
        assert (
            len(mul_t) == 1
            and mul_t[0].variable() is v
            and mul_t[0].coefficient() == 20
            and mul.constant() == 10
        )

    with pytest.raises(TypeError):
        e * v  # type: ignore


def test_expression_div(
    expressions: Tuple[Expression, Expression, Term, Term, Variable, Variable]
):
    """Test expression divisions."""
    e, _, _, _, v, v2 = expressions

    div = e / 2
    assert isinstance(div, Expression)
    div_t = div.terms()
    assert (
        len(div_t) == 1
        and div_t[0].variable() is v
        and div_t[0].coefficient() == 5
        and div.constant() == 2.5
    )

    with pytest.raises(TypeError):
        e / v2  # type: ignore

    with pytest.raises(ZeroDivisionError):
        e / 0


def test_expression_addition(
    expressions: Tuple[Expression, Expression, Term, Term, Variable, Variable]
):
    """Test expressions additions."""
    e, e2, _, t2, v, v2 = expressions

    for add in (e + 2, 2.0 + e):
        assert isinstance(add, Expression)
        assert add.constant() == 7
        terms = add.terms()
        assert (
            len(terms) == 1
            and terms[0].variable() is v
            and terms[0].coefficient() == 10
        )

    add2 = e + v2
    assert isinstance(add2, Expression)
    assert add2.constant() == 5
    terms = add2.terms()
    assert (
        len(terms) == 2
        and terms[0].variable() is v
        and terms[0].coefficient() == 10
        and terms[1].variable() is v2
        and terms[1].coefficient() == 1
    )

    add3 = e + t2
    assert isinstance(add3, Expression)
    assert add3.constant() == 5
    terms = add3.terms()
    assert (
        len(terms) == 2
        and terms[0].variable() is v
        and terms[0].coefficient() == 10
        and terms[1].variable() is v2
        and terms[1].coefficient() == 1
    )

    add4 = e + e2
    assert isinstance(add4, Expression)
    assert add4.constant() == -5
    terms = add4.terms()
    assert (
        len(terms) == 2
        and terms[0].variable() is v
        and terms[0].coefficient() == 10
        and terms[1].variable() is v2
        and terms[1].coefficient() == 1
    )


def test_expressions_substraction(
    expressions: Tuple[Expression, Expression, Term, Term, Variable, Variable]
):
    """Test expression substraction."""
    e, e2, _, t2, v, v2 = expressions

    for sub, diff in zip((e - 2, 2.0 - e), (3, -3)):
        assert isinstance(sub, Expression)
        assert sub.constant() == diff
        terms = sub.terms()
        assert (
            len(terms) == 1
            and terms[0].variable() is v
            and terms[0].coefficient() == math.copysign(10, diff)
        )

    for sub2, diff in zip((e - v2, v2 - e), (5, -5)):
        assert isinstance(sub2, Expression)
        assert sub2.constant() == diff
        terms = sub2.terms()
        assert (
            len(terms) == 2
            and terms[0].variable() is v
            and terms[0].coefficient() == math.copysign(10, diff)
            and terms[1].variable() is v2
            and terms[1].coefficient() == -math.copysign(1, diff)
        )

    for sub3, diff in zip((e - t2, t2 - e), (5, -5)):
        assert isinstance(sub3, Expression)
        assert sub3.constant() == diff
        terms = sub3.terms()
        assert (
            len(terms) == 2
            and terms[0].variable() is v
            and terms[0].coefficient() == math.copysign(10, diff)
            and terms[1].variable() is v2
            and terms[1].coefficient() == -math.copysign(1, diff)
        )

    sub4 = e - e2
    assert isinstance(sub3, Expression)
    assert sub4.constant() == 15
    terms = sub4.terms()
    assert (
        len(terms) == 2
        and terms[0].variable() is v
        and terms[0].coefficient() == 10
        and terms[1].variable() is v2
        and terms[1].coefficient() == -1
    )


@pytest.mark.parametrize(
    "op, symbol",
    [
        (operator.le, "<="),
        (operator.eq, "=="),
        (operator.ge, ">="),
        (operator.lt, None),
        (operator.ne, None),
        (operator.gt, None),
    ],
)
def test_expression_rich_compare_operations(op, symbol) -> None:
    """Test using comparison on variables."""
    v1 = Variable("foo")
    v2 = Variable("bar")
    t1 = Term(v1, 10)
    e1 = t1 + 5
    e2 = v2 - 10

    if symbol is not None:
        c = op(e1, e2)
        assert isinstance(c, Constraint)
        e = c.expression()
        t = e.terms()
        assert len(t) == 2
        if t[0].variable() is not v1:
            t = (t[1], t[0])
        assert (
            t[0].variable() is v1
            and t[0].coefficient() == 10
            and t[1].variable() is v2
            and t[1].coefficient() == -1
        )
        assert e.constant() == 15
        assert c.op() == symbol and c.strength() == strength.required

    else:
        with pytest.raises(TypeError) as excinfo:
            op(e1, e2)
        assert "kiwisolver.Expression" in excinfo.exconly()



# --------------------------------------------------------------------------------------
# Copyright (c) 2013-2022, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# --------------------------------------------------------------------------------------
import os

from setuptools import Extension, setup

try:
    from cppy import CppyBuildExt
except ImportError as e:
    raise RuntimeError(
        "Missing setup required dependencies: cppy. "
        "Installing through pip as recommended ensure one never hits this issue."
    ) from e

# Before releasing the version needs to be updated in kiwi/version.h, if the changes
# are not limited to the solver.

# Use the env var KIWI_DISABLE_FH4 to disable linking against VCRUNTIME140_1.dll

if "KIWI_DISABLE_FH4" in os.environ:
    os.environ.setdefault("CPPY_DISABLE_FH4", "1")

ext_modules = [
    Extension(
        "kiwisolver._cext",
        [
            "py/src/kiwisolver.cpp",
            "py/src/constraint.cpp",
            "py/src/expression.cpp",
            "py/src/solver.cpp",
            "py/src/strength.cpp",
            "py/src/term.cpp",
            "py/src/variable.cpp",
        ],
        include_dirs=["."],
        language="c++",
    ),
]


setup(
    ext_modules=ext_modules,
    cmdclass={"build_ext": CppyBuildExt},
)



# --------------------------------------------------------------------------------------
# Copyright (c) 2014-2022, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# --------------------------------------------------------------------------------------
import math
import operator
import sys

import pytest

from kiwisolver import Constraint, Expression, Term, Variable, strength


def test_variable_methods() -> None:
    """Test the variable modification methods."""
    v = Variable()
    assert v.name() == ""
    v.setName("γ")
    assert v.name() == "γ"
    v.setName("foo")
    assert v.name() == "foo"
    with pytest.raises(TypeError):
        v.setName(1)  # type: ignore
    if sys.version_info >= (3,):
        with pytest.raises(TypeError):
            v.setName(b"r")  # type: ignore

    assert v.value() == 0.0

    assert v.context() is None
    ctx = object()
    v.setContext(ctx)
    assert v.context() is ctx

    assert str(v) == "foo"

    with pytest.raises(TypeError):
        Variable(1)  # type: ignore


def test_variable_neg() -> None:
    """Test neg on a variable."""
    v = Variable("foo")

    neg = -v
    assert isinstance(neg, Term)
    assert neg.variable() is v and neg.coefficient() == -1


def test_variable_mul() -> None:
    """Test variable multiplications."""
    v = Variable("foo")
    v2 = Variable("bar")

    for mul in (v * 2.0, 2 * v):
        assert isinstance(mul, Term)
        assert mul.variable() is v and mul.coefficient() == 2

    with pytest.raises(TypeError):
        v * v2  # type: ignore


def test_variable_division() -> None:
    """Test variable divisions."""
    v = Variable("foo")
    v2 = Variable("bar")

    div = v / 2.0
    assert isinstance(div, Term)
    assert div.variable() is v and div.coefficient() == 0.5

    with pytest.raises(TypeError):
        v / v2  # type: ignore

    with pytest.raises(ZeroDivisionError):
        v / 0


def test_variable_addition() -> None:
    """Test variable additions."""
    v = Variable("foo")
    v2 = Variable("bar")

    for add in (v + 2, 2.0 + v):
        assert isinstance(add, Expression)
        assert add.constant() == 2
        terms = add.terms()
        assert (
            len(terms) == 1 and terms[0].variable() is v and terms[0].coefficient() == 1
        )

    add2 = v + v2
    assert isinstance(add2, Expression)
    assert add2.constant() == 0
    terms = add2.terms()
    assert (
        len(terms) == 2
        and terms[0].variable() is v
        and terms[0].coefficient() == 1
        and terms[1].variable() is v2
        and terms[1].coefficient() == 1
    )

    with pytest.raises(TypeError):
        v + ""  # type: ignore


def test_variable_sub() -> None:
    """Test variable substractions."""
    v = Variable("foo")
    v2 = Variable("bar")

    for sub, diff in zip((v - 2, 2 - v), (-2, 2)):
        assert isinstance(sub, Expression)
        assert sub.constant() == diff
        terms = sub.terms()
        assert (
            len(terms) == 1
            and terms[0].variable() is v
            and terms[0].coefficient() == -math.copysign(1, diff)
        )

    sub2 = v - v2
    assert isinstance(sub2, Expression)
    assert sub2.constant() == 0
    terms = sub2.terms()
    assert (
        len(terms) == 2
        and terms[0].variable() is v
        and terms[0].coefficient() == 1
        and terms[1].variable() is v2
        and terms[1].coefficient() == -1
    )


def test_variable_rich_compare_operations() -> None:
    """Test using comparison on variables."""
    v = Variable("foo")
    v2 = Variable("γ")

    for op, symbol in ((operator.le, "<="), (operator.eq, "=="), (operator.ge, ">=")):
        c = op(v, v2 + 1)
        assert isinstance(c, Constraint)
        e = c.expression()
        t = e.terms()
        assert len(t) == 2
        if t[0].variable() is not v:
            t = (t[1], t[0])
        assert (
            t[0].variable() is v
            and t[0].coefficient() == 1
            and t[1].variable() is v2
            and t[1].coefficient() == -1
        )
        assert e.constant() == -1
        assert c.op() == symbol and c.strength() == strength.required

    for op in (operator.lt, operator.ne, operator.gt):
        with pytest.raises(TypeError):
            op(v, v2)
